import re
import time
import traceback
from threading import Thread
from typing import List, Set, Type, Optional, Tuple, Generator

from bauh.api.abstract.controller import SoftwareManager, SearchResult, ApplicationContext, UpgradeRequirements, \
    TransactionResult, SoftwareAction
from bauh.api.abstract.disk import DiskCacheLoader
from bauh.api.abstract.handler import ProcessWatcher, TaskManager
from bauh.api.abstract.model import SoftwarePackage, PackageHistory, PackageUpdate, PackageSuggestion, \
    SuggestionPriority, PackageStatus
from bauh.api.abstract.view import SingleSelectComponent, SelectViewType, InputOption, ViewComponent, PanelComponent, \
    FormComponent, TextInputComponent
from bauh.api.exception import NoInternetException
from bauh.commons.boot import CreateConfigFile
from bauh.commons.category import CategoriesDownloader
from bauh.commons.html import bold
from bauh.commons.system import SystemProcess, ProcessHandler, new_root_subprocess
from bauh.commons.view_utils import new_select, get_human_size_str
from bauh.gems.snap import snap, URL_CATEGORIES_FILE, CATEGORIES_FILE_PATH, SUGGESTIONS_FILE, \
    get_icon_path, snapd
from bauh.gems.snap.config import SnapConfigManager
from bauh.gems.snap.model import SnapApplication
from bauh.gems.snap.snapd import SnapdClient

RE_AVAILABLE_CHANNELS = re.compile(re.compile(r'(\w+)\s+(snap install.+)'))


class SnapManager(SoftwareManager):

    def __init__(self, context: ApplicationContext):
        super(SnapManager, self).__init__(context=context)
        self.i18n = context.i18n
        self.api_cache = context.cache_factory.new()
        context.disk_loader_factory.map(SnapApplication, self.api_cache)
        self.enabled = True
        self.http_client = context.http_client
        self.logger = context.logger
        self.ubuntu_distro = context.distro == 'ubuntu'
        self.categories = {}
        self.suggestions_cache = context.cache_factory.new()
        self.info_path = None
        self.configman = SnapConfigManager()

    def _fill_categories(self, app: SnapApplication):
        categories = self.categories.get(app.name.lower())

        if categories:
            app.categories = categories

        if not app.is_application():
            categories = app.categories

            if categories is None:
                categories = []
                app.categories = categories

            if 'runtime' not in categories:
                categories.append('runtime')

    def search(self, words: str, disk_loader: DiskCacheLoader, limit: int = -1, is_url: bool = False) -> SearchResult:
        if is_url or (not snap.is_installed() and not snapd.is_running()):
            return SearchResult([], [], 0)

        snapd_client = SnapdClient(self.logger)
        apps_found = snapd_client.query(words)

        res = SearchResult([], [], 0)

        if apps_found:
            installed = self.read_installed(disk_loader).installed

            for app_json in apps_found:
                already_installed = None

                if installed:
                    already_installed = [i for i in installed if i.id == app_json.get('id')]
                    already_installed = already_installed[0] if already_installed else None

                if already_installed:
                    res.installed.append(already_installed)
                else:
                    res.new.append(self._map_to_app(app_json, installed=False))

        res.total = len(res.installed) + len(res.new)
        return res

    def read_installed(self, disk_loader: DiskCacheLoader, limit: int = -1, only_apps: bool = False, pkg_types: Set[Type[SoftwarePackage]] = None, internet_available: bool = None) -> SearchResult:
        if snap.is_installed() and snapd.is_running():
            snapd_client = SnapdClient(self.logger)
            app_names = {a['snap'] for a in snapd_client.list_only_apps()}
            installed = [self._map_to_app(app_json=appjson,
                                          installed=True,
                                          disk_loader=disk_loader,
                                          is_application=app_names and appjson['name'] in app_names) for appjson in snapd_client.list_all_snaps()]
            return SearchResult(installed, None, len(installed))
        else:
            return SearchResult([], None, 0)

    def downgrade(self, pkg: SnapApplication, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        if not snap.is_installed():
            watcher.print("'snap' seems not to be installed")
            return False
        if not snapd.is_running():
            watcher.print("'snapd' seems not to be running")
            return False

        return ProcessHandler(watcher).handle_simple(snap.downgrade_and_stream(pkg.name, root_password))[0]

    def upgrade(self, requirements: UpgradeRequirements, root_password: Optional[str], watcher: ProcessWatcher) -> SystemProcess:
        raise Exception(f"'upgrade' is not supported by {SnapManager.__class__.__name__}")

    def uninstall(self, pkg: SnapApplication, root_password: Optional[str], watcher: ProcessWatcher, disk_loader: DiskCacheLoader) -> TransactionResult:
        if snap.is_installed() and snapd.is_running():
            uninstalled = ProcessHandler(watcher).handle_simple(snap.uninstall_and_stream(pkg.name, root_password))[0]

            if uninstalled:
                if self.suggestions_cache:
                    self.suggestions_cache.delete(pkg.name)

                return TransactionResult(success=True, installed=None, removed=[pkg])

        return TransactionResult.fail()

    def get_managed_types(self) -> Set[Type[SoftwarePackage]]:
        return {SnapApplication}

    def clean_cache_for(self, pkg: SnapApplication):
        super(SnapManager, self).clean_cache_for(pkg)
        self.api_cache.delete(pkg.id)

    def get_info(self, pkg: SnapApplication) -> dict:
        info = {
            'description': pkg.description,
            'developer': pkg.developer,
            'license': pkg.license,
            'contact': pkg.contact,
            'snap-id': pkg.id,
            'name': pkg.name,
            'publisher': pkg.publisher,
            'revision': pkg.rev,
            'tracking': pkg.tracking,
            'channel': pkg.channel,
            'type': pkg.type
        }

        if pkg.installed:
            commands = [*{c['name'] for c in SnapdClient(self.logger).list_commands(pkg.name)}]
            commands.sort()
            info['commands'] = commands

            if pkg.installed_size:
                info['installed_size']: get_human_size_str(pkg.installed_size)
        elif pkg.download_size:
            info['download_size'] = get_human_size_str(pkg.download_size)

        return info

    def get_history(self, pkg: SnapApplication) -> PackageHistory:
        raise Exception(f"'get_history' is not supported by {pkg.__class__.__name__}")

    def install(self, pkg: SnapApplication, root_password: Optional[str], disk_loader: DiskCacheLoader, watcher: ProcessWatcher) -> TransactionResult:
        # retrieving all installed so it will be possible to know the additional installed runtimes after the operation succeeds
        if not snap.is_installed():
            watcher.print("'snap' seems not to be installed")
            return TransactionResult.fail()

        if not snapd.is_running():
            watcher.print("'snapd' seems not to be running")
            return TransactionResult.fail()

        installed_names = {s['name'] for s in SnapdClient(self.logger).list_all_snaps()}

        client = SnapdClient(self.logger)
        snap_config = self.configman.get_config()

        try:
            channel = self._request_channel_installation(pkg=pkg, snap_config=snap_config, snapd_client=client, watcher=watcher)
            pkg.channel = channel
        except:
            watcher.print('Aborted by user')
            return TransactionResult.fail()

        res, output = ProcessHandler(watcher).handle_simple(snap.install_and_stream(app_name=pkg.name,
                                                                                    confinement=pkg.confinement,
                                                                                    root_password=root_password,
                                                                                    channel=channel))

        if 'error:' in output:
            res = False
            if 'not available on stable' in output:
                channels = RE_AVAILABLE_CHANNELS.findall(output)

                if channels:
                    opts = [InputOption(label=c[0], value=c[1]) for c in channels]
                    channel_select = SingleSelectComponent(type_=SelectViewType.RADIO, label='', options=opts, default_option=opts[0])
                    body = f"<p>{self.i18n['snap.install.available_channels.message'].format(bold(self.i18n['stable']), bold(pkg.name))}.</p>"
                    body += f"<p>{self.i18n['snap.install.available_channels.help']}:</p>"

                    if watcher.request_confirmation(title=self.i18n['snap.install.available_channels.title'],
                                                    body=body,
                                                    components=[channel_select],
                                                    confirmation_label=self.i18n['continue'],
                                                    deny_label=self.i18n['cancel']):
                        self.logger.info(f"Installing '{pkg.name}' with the custom command '{channel_select.value}'")
                        res = ProcessHandler(watcher).handle(SystemProcess(new_root_subprocess(channel_select.value.value.split(' '), root_password=root_password)))
                        return self._gen_installation_response(success=res, pkg=pkg,
                                                               installed=installed_names, disk_loader=disk_loader)
                else:
                    self.logger.error(f"Could not find available channels in the installation output: {output}")

        return self._gen_installation_response(success=res, pkg=pkg, installed=installed_names, disk_loader=disk_loader)

    def _gen_installation_response(self, success: bool, pkg: SnapApplication, installed: Set[str], disk_loader: DiskCacheLoader):
        if success:
            new_installed = []
            try:
                net_available = self.context.internet_checker.is_available()
                current_installed = self.read_installed(disk_loader=disk_loader, internet_available=net_available).installed
            except:
                new_installed = [pkg]
                traceback.print_exc()
                current_installed = None

            if current_installed:
                for p in current_installed:
                    if p.name == pkg.name or (not installed or p.name not in installed):
                        new_installed.append(p)

            return TransactionResult(success=success, installed=new_installed, removed=[])
        else:
            return TransactionResult.fail()

    def is_enabled(self) -> bool:
        return self.enabled

    def set_enabled(self, enabled: bool):
        self.enabled = enabled

    def can_work(self) -> Tuple[bool, Optional[str]]:
        return (True, None) if snap.is_installed() else (False, self.i18n['missing_dep'].format(dep=bold('snap')))

    def requires_root(self, action: SoftwareAction, pkg: SnapApplication) -> bool:
        return action not in (SoftwareAction.PREPARE, SoftwareAction.SEARCH)

    def refresh(self, pkg: SnapApplication, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        return ProcessHandler(watcher).handle_simple(snap.refresh_and_stream(pkg.name, root_password))[0]

    def change_channel(self, pkg: SnapApplication, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        if not self.context.internet_checker.is_available():
            raise NoInternetException()

        try:
            channel = self._request_channel_installation(pkg=pkg,
                                                         snap_config=None,
                                                         snapd_client=SnapdClient(self.logger),
                                                         watcher=watcher,
                                                         exclude_current=True)

            if not channel:
                watcher.show_message(title=self.i18n['snap.action.channel.label'],
                                     body=self.i18n['snap.action.channel.error.no_channel'])
                return False

            return ProcessHandler(watcher).handle_simple(snap.refresh_and_stream(app_name=pkg.name,
                                                                                 root_password=root_password,
                                                                                 channel=channel))[0]
        except:
            return False

    def _start_category_task(self, taskman: TaskManager, create_config: CreateConfigFile, downloader: CategoriesDownloader):
        if taskman:
            taskman.update_progress('snap_cats', 0, self.i18n['task.waiting_task'].format(bold(create_config.task_name)))
            create_config.join()

            categories_exp = create_config.config['categories_exp']
            downloader.expiration = categories_exp if isinstance(categories_exp, int) else None
            taskman.update_progress('snap_cats', 1, None)

    def _finish_category_task(self, taskman: TaskManager):
        if taskman:
            taskman.update_progress('snap_cats', 100, None)
            taskman.finish_task('snap_cats')

    def prepare(self, task_manager: TaskManager, root_password: Optional[str], internet_available: bool):
        create_config = CreateConfigFile(taskman=task_manager, configman=self.configman, i18n=self.i18n,
                                         task_icon_path=get_icon_path(), logger=self.logger)
        create_config.start()

        task_manager.register_task('snap_cats', self.i18n['task.download_categories'], get_icon_path())
        category_downloader = CategoriesDownloader(id_='snap', manager=self, http_client=self.http_client,
                                                   logger=self.logger,
                                                   url_categories_file=URL_CATEGORIES_FILE,
                                                   categories_path=CATEGORIES_FILE_PATH,
                                                   internet_connection=internet_available,
                                                   internet_checker=self.context.internet_checker,
                                                   after=lambda: self._finish_category_task(task_manager))
        category_downloader.before = lambda: self._start_category_task(task_manager, create_config, category_downloader)
        category_downloader.start()

    def list_updates(self, internet_available: bool) -> List[PackageUpdate]:
        pass

    def list_warnings(self, internet_available: bool) -> Optional[List[str]]:
        if not snapd.is_running():
            snap_bold = bold('Snap')
            return [self.i18n['snap.notification.snapd_unavailable'].format(bold('snapd'), snap_bold),
                    self.i18n['snap.notification.snap.disable'].format(snap_bold,
                                                                       bold(f"{self.i18n['settings'].capitalize()} > {self.i18n['core.config.tab.types']}"))]
        elif internet_available:
            available, output = snap.is_api_available()

            if not available:
                self.logger.warning(f'It seems Snap API is not available. Search output: {output}')
                return [self.i18n['snap.notifications.api.unavailable'].format(bold('Snaps'), bold('Snap'))]

    def _fill_suggestion(self, name: str, priority: SuggestionPriority, snapd_client: SnapdClient, out: List[PackageSuggestion]):
        res = snapd_client.find_by_name(name)

        if res:
            if len(res) == 1:
                app_json = res[0]
            else:
                jsons_found = [p for p in res if p['name'] == name]
                app_json = jsons_found[0] if jsons_found else None

            if app_json:
                sug = PackageSuggestion(self._map_to_app(app_json, False), priority)
                self.suggestions_cache.add(name, sug)
                out.append(sug)
                return

        self.logger.warning(f"Could not retrieve suggestion '{name}'")

    def _map_to_app(self, app_json: dict, installed: bool, disk_loader: Optional[DiskCacheLoader] = None, is_application: bool = False) -> SnapApplication:
        app = SnapApplication(id=app_json.get('id'),
                              name=app_json.get('name'),
                              license=app_json.get('license'),
                              version=app_json.get('version'),
                              latest_version=app_json.get('version'),
                              description=app_json.get('description', app_json.get('summary')),
                              installed=installed,
                              rev=app_json.get('revision'),
                              publisher=app_json['publisher'].get('display-name', app_json['publisher'].get('username')),
                              verified_publisher=app_json['publisher'].get('validation') == 'verified',
                              icon_url=app_json.get('icon'),
                              screenshots={m['url'] for m in app_json.get('media', []) if m['type'] == 'screenshot'},
                              download_size=app_json.get('download-size'),
                              channel=app_json.get('channel'),
                              confinement=app_json.get('confinement'),
                              app_type=app_json.get('type'),
                              app=is_application,
                              installed_size=app_json.get('installed-size'))

        if disk_loader and app.installed:
            disk_loader.fill(app)

        self._fill_categories(app)

        app.status = PackageStatus.READY
        return app

    def list_suggestions(self, limit: int, filter_installed: bool) -> List[PackageSuggestion]:
        res = []

        if snapd.is_running():
            self.logger.info(f'Downloading suggestions file {SUGGESTIONS_FILE}')
            file = self.http_client.get(SUGGESTIONS_FILE)

            if not file or not file.text:
                self.logger.warning(f"No suggestion found in {SUGGESTIONS_FILE}")
                return res
            else:
                self.logger.info('Mapping suggestions')

                suggestions, threads = [], []
                snapd_client = SnapdClient(self.logger)
                installed = {s['name'].lower() for s in snapd_client.list_all_snaps()}

                for l in file.text.split('\n'):
                    if l:
                        if limit <= 0 or len(suggestions) < limit:
                            sug = l.strip().split('=')
                            name = sug[1]

                            if not installed or name not in installed:
                                cached_sug = self.suggestions_cache.get(name)

                                if cached_sug:
                                    res.append(cached_sug)
                                else:
                                    t = Thread(target=self._fill_suggestion, args=(name, SuggestionPriority(int(sug[0])), snapd_client, res))
                                    t.start()
                                    threads.append(t)
                                    time.sleep(0.001)  # to avoid being blocked
                        else:
                            break

                for t in threads:
                    t.join()

                res.sort(key=lambda s: s.priority.value, reverse=True)
        return res

    def is_default_enabled(self) -> bool:
        return True

    def launch(self, pkg: SnapApplication):
        commands = SnapdClient(self.logger).list_commands(pkg.name)

        if commands:
            if len(commands) == 1:
                cmd = commands[0]['name']
            else:
                desktop_cmd = [c for c in commands if 'desktop-file' in c]

                if desktop_cmd:
                    cmd = desktop_cmd[0]['name']
                else:
                    cmd = commands[0]['name']

            self.logger.info(f"Running '{pkg.name}': {cmd}")
            snap.run(cmd)

    def get_screenshots(self, pkg: SnapApplication) -> Generator[str, None, None]:
        if pkg.screenshots:
            yield from pkg.screenshots

    def get_settings(self, screen_width: int, screen_height: int) -> Optional[ViewComponent]:
        snap_config = self.configman.get_config()
        max_width = 200

        install_channel = new_select(label=self.i18n['snap.config.install_channel'],
                                     opts=[(self.i18n['yes'].capitalize(), True, None),
                                           (self.i18n['no'].capitalize(), False, None)],
                                     value=bool(snap_config['install_channel']),
                                     id_='snap_install_channel',
                                     max_width=max_width,
                                     tip=self.i18n['snap.config.install_channel.tip'])

        categories_exp = TextInputComponent(id_='snap_cat_exp',
                                            value=snap_config['categories_exp'] if isinstance(snap_config['categories_exp'], int) else '',
                                            max_width=max_width,
                                            only_int=True,
                                            label=self.i18n['snap.config.categories_exp'],
                                            tooltip=self.i18n['snap.config.categories_exp.tip'])

        return PanelComponent([FormComponent([install_channel, categories_exp], self.i18n['installation'].capitalize())])

    def save_settings(self, component: ViewComponent) -> Tuple[bool, Optional[List[str]]]:
        snap_config = self.configman.get_config()

        panel = component.components[0]
        snap_config['install_channel'] = panel.get_component('snap_install_channel').get_selected()
        snap_config['categories_exp'] = panel.get_component('snap_cat_exp').get_int_value()

        try:
            self.configman.save_config(snap_config)
            return True, None
        except:
            return False, [traceback.format_exc()]

    def _request_channel_installation(self, pkg: SnapApplication, snap_config: Optional[dict], snapd_client: SnapdClient, watcher: ProcessWatcher, exclude_current: bool = False) -> Optional[str]:
        if snap_config is None or snap_config['install_channel']:
            try:
                data = [r for r in snapd_client.find_by_name(pkg.name) if r['name'] == pkg.name]
            except:
                self.logger.warning(f"snapd client could not retrieve channels for '{pkg.name}'")
                return

            if not data:
                self.logger.warning(f"snapd client could find a match for name '{pkg.name}' when retrieving its channels")
            else:
                if not data[0].get('channels'):
                    self.logger.info(f"No channel available for '{pkg.name}'. Skipping selection.")
                else:
                    if pkg.channel:
                        current_channel = pkg.channel if '/' in pkg.channel else f'latest/{pkg.channel}'
                    else:
                        current_channel = f"latest/{data[0].get('channel', 'stable')}"

                    opts = []
                    def_opt = None
                    for channel in sorted(data[0]['channels'].keys()):
                        if exclude_current:
                            if channel != current_channel:
                                opts.append(InputOption(label=channel, value=channel))
                        else:
                            op = InputOption(label=channel, value=channel)
                            opts.append(op)

                            if not def_opt and channel == current_channel:
                                def_opt = op

                    if not opts:
                        self.logger.info(f"No different channel available for '{pkg.name}'. Skipping selection.")
                        return

                    select = SingleSelectComponent(label='',
                                                   options=opts,
                                                   default_option=def_opt if def_opt else opts[0],
                                                   type_=SelectViewType.RADIO)

                    if not watcher.request_confirmation(title=self.i18n['snap.install.available_channels.title'],
                                                        body=self.i18n['snap.install.channel.body'] + ':',
                                                        components=[select],
                                                        confirmation_label=self.i18n['proceed'].capitalize(),
                                                        deny_label=self.i18n['cancel'].capitalize()):
                        raise Exception('aborted')
                    else:
                        return select.get_selected()



import faulthandler
import os
import sys
import traceback

import urllib3
from PyQt5.QtCore import QCoreApplication, Qt

from bauh import __app_name__, app_args
from bauh.view.core.config import CoreConfigManager
from bauh.view.util import logs


def main(tray: bool = False):
    if not os.getenv('PYTHONUNBUFFERED'):
        os.environ['PYTHONUNBUFFERED'] = '1'

    if not os.getenv('XDG_RUNTIME_DIR'):
        os.environ['XDG_RUNTIME_DIR'] = f'/run/user/{os.getuid()}'

    faulthandler.enable()
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    args = app_args.read()

    logger = logs.new_logger(__app_name__, bool(args.logs))

    if args.offline:
        logger.warning("offline mode activated")

    app_config = CoreConfigManager().get_config()

    if bool(app_config['ui']['auto_scale']):
        os.environ['QT_AUTO_SCREEN_SCALE_FACTOR'] = '1'
        logger.info("Auto screen scale factor activated")

    try:
        scale_factor = float(app_config['ui']['scale_factor'])
        os.environ['QT_SCALE_FACTOR'] = str(scale_factor)
        logger.info("Scale factor set to {}".format(scale_factor))
    except:
        traceback.print_exc()

    if bool(app_config['ui']['hdpi']):
        logger.info("HDPI settings activated")
        QCoreApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
        QCoreApplication.setAttribute(Qt.AA_EnableHighDpiScaling)

    if tray or bool(args.tray):
        from bauh.tray import new_tray_icon
        app, widget = new_tray_icon(app_config, logger)
    else:
        from bauh.manage import new_manage_panel
        app, widget = new_manage_panel(args, app_config, logger)

    widget.show()
    sys.exit(app.exec_())


def tray():
    main(tray=True)


if __name__ == '__main__':
    main()



from collections.abc import Iterable

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QIcon, QCursor
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QGroupBox, \
    QLineEdit, QLabel, QGridLayout, QPushButton, QPlainTextEdit, QScrollArea, QFrame, QWidget, QSizePolicy, \
    QHBoxLayout

from bauh.api.abstract.cache import MemoryCache
from bauh.view.qt.components import new_spacer
from bauh.view.util.translation import I18n

IGNORED_ATTRS = {'name', '__app__'}


class InfoDialog(QDialog):

    def __init__(self, pkg_info: dict, icon_cache: MemoryCache, i18n: I18n, screen_size: QSize):
        super(InfoDialog, self).__init__()
        self.setWindowTitle(str(pkg_info['__app__']))
        self.screen_size = screen_size
        self.i18n = i18n
        layout = QVBoxLayout()
        self.setLayout(layout)

        scroll = QScrollArea(self)
        scroll.setFrameShape(QFrame.NoFrame)
        scroll.setWidgetResizable(True)
        comps_container = QWidget()
        comps_container.setObjectName('root_container')
        comps_container.setLayout(QVBoxLayout())
        scroll.setWidget(comps_container)

        # shows complete field string
        self.text_field = QPlainTextEdit()
        self.text_field.setObjectName('full_field')
        self.text_field.setReadOnly(True)
        comps_container.layout().addWidget(self.text_field)
        self.text_field.hide()

        self.gbox_info = QGroupBox()
        self.gbox_info.setObjectName('fields')
        self.gbox_info.setLayout(QGridLayout())

        comps_container.layout().addWidget(self.gbox_info)

        # THERE ARE CRASHES WITH SOME RARE ICONS ( like insomnia ). IT CAN BE A QT BUG. IN THE MEANTIME, ONLY THE TYPE ICON WILL BE RENDERED
        #
        # icon_data = icon_cache.get(app['__app__'].model.icon_url)
        #
        # if icon_data and icon_data.get('icon'):
        #     self.setWindowIcon(icon_data.get('icon'))
        self.setWindowIcon(QIcon(pkg_info['__app__'].model.get_type_icon_path()))

        for idx, attr in enumerate(sorted(pkg_info.keys())):
            if attr not in IGNORED_ATTRS and pkg_info[attr] is not None:
                i18n_key = pkg_info['__app__'].model.gem_name + '.info.' + attr.lower()
                val = pkg_info[attr]

                if not isinstance(val, str) and isinstance(pkg_info[attr], Iterable):
                    val = ' '.join([str(e).strip() for e in pkg_info[attr] if e])
                    show_val = '\n'.join(['* ' + str(e).strip() for e in pkg_info[attr] if e])
                else:
                    val = str(pkg_info[attr]).strip()
                    show_val = val

                i18n_val = i18n.get('{}.{}'.format(i18n_key, val.lower()))

                if i18n_val:
                    val = i18n_val
                    show_val = val

                text = QLineEdit()
                text.setObjectName('field_value')
                text.setToolTip(show_val)
                text.setText(val)
                text.setCursorPosition(0)
                text.setReadOnly(True)

                label = QLabel(i18n.get(i18n_key, i18n.get(attr.lower(), attr)).capitalize())
                label.setObjectName('field_name')

                self.gbox_info.layout().addWidget(label, idx, 0)
                self.gbox_info.layout().addWidget(text, idx, 1)
                self._gen_show_button(idx, show_val)

        layout.addWidget(scroll)

        lower_container = QWidget()
        lower_container.setObjectName('lower_container')
        lower_container.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        lower_container.setLayout(QHBoxLayout())

        self.bt_back = QPushButton('< {}'.format(self.i18n['back'].capitalize()))
        self.bt_back.setObjectName('back')
        self.bt_back.setVisible(False)
        self.bt_back.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_back.clicked.connect(self.back_to_info)

        lower_container.layout().addWidget(self.bt_back)
        lower_container.layout().addWidget(new_spacer())

        self.bt_close = QPushButton(self.i18n['close'].capitalize())
        self.bt_close.setObjectName('close')
        self.bt_close.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_close.clicked.connect(self.close)

        lower_container.layout().addWidget(self.bt_close)
        layout.addWidget(lower_container)
        self.setMinimumWidth(int(self.gbox_info.sizeHint().width() * 1.2))
        self.setMaximumHeight(int(screen_size.height() * 0.8))
        self.adjustSize()

    def _gen_show_button(self, idx: int, val):

        def show_full_field():
            self.gbox_info.hide()
            self.bt_back.setVisible(True)
            self.text_field.show()
            self.text_field.setPlainText(val)

        bt_full_field = QPushButton(self.i18n['show'].capitalize())
        bt_full_field.setObjectName('show')
        bt_full_field.setCursor(QCursor(Qt.PointingHandCursor))
        bt_full_field.clicked.connect(show_full_field)
        self.gbox_info.layout().addWidget(bt_full_field, idx, 2)

    def back_to_info(self):
        self.text_field.setPlainText("")
        self.text_field.hide()
        self.gbox_info.show()
        self.bt_back.setVisible(False)



import warnings
from unittest import TestCase
from unittest.mock import Mock

from bauh.gems.arch.mapper import AURDataMapper
from bauh.gems.arch.model import ArchPackage


class ArchDataMapperTest(TestCase):

    @classmethod
    def setUpClass(cls):
        warnings.filterwarnings('ignore', category=DeprecationWarning)

    def test_check_version_update(self):
        self.assertTrue(AURDataMapper.check_version_update('1.0.0-1', '1.0.0-2'))
        self.assertFalse(AURDataMapper.check_version_update('1.0.0-2', '1.0.0-1'))
        self.assertTrue(AURDataMapper.check_version_update('1.0.0-5', '1.0.1-1'))
        self.assertFalse(AURDataMapper.check_version_update('1.0.1-1', '1.0.0-1'))
        self.assertTrue(AURDataMapper.check_version_update('1.0.5-5', '1.1.0-2'))
        self.assertFalse(AURDataMapper.check_version_update('1.1.0-2', '1.0.5-5'))
        self.assertTrue(AURDataMapper.check_version_update('1.5.0-2', '1.5.1-1'))
        self.assertFalse(AURDataMapper.check_version_update('1.5.1-1', '1.5.0-2'))
        self.assertTrue(AURDataMapper.check_version_update('1.5.1-1', '1.5.1-2'))
        self.assertTrue(AURDataMapper.check_version_update('1.5.1-1', '2.0.0-1'))
        self.assertFalse(AURDataMapper.check_version_update('2.0.0-1', '1.5.1-1'))
        self.assertTrue(AURDataMapper.check_version_update('77.0.3865.90-1', '77.0.3865.120-1'))
        self.assertTrue(AURDataMapper.check_version_update('77.0.3865.90-1', '77.0.3865.90-2'))
        self.assertFalse(AURDataMapper.check_version_update('77.0.3865.900-1', '77.0.3865.120-1'))
        self.assertTrue(AURDataMapper.check_version_update('77.0.3865.120-1', '77.0.3865.900-1'))
        self.assertFalse(AURDataMapper.check_version_update('77.0.3865.120-1', '77.0.3865.90-1'))
        self.assertTrue(AURDataMapper.check_version_update('77.0.3865.a-1', '77.0.3865.b-1'))
        self.assertFalse(AURDataMapper.check_version_update('77.0.b.0-1', '77.0.a.1-1'))
        self.assertFalse(AURDataMapper.check_version_update('r25.e22697c-1', 'r8.19fe011-1'))
        self.assertTrue(AURDataMapper.check_version_update('0.9.7.RC-9', '0.9.7.RC-10'))
        self.assertFalse(AURDataMapper.check_version_update('1.1.0.r11.caacf30-1', 'r65.4c7144a-1'))
        self.assertFalse(AURDataMapper.check_version_update('1.2.16.r688.8b2c199-1', 'r2105.e91f0e9-3'))

    def test_check_version_update__versions_with_epics(self):
        self.assertTrue(AURDataMapper.check_version_update('1.2-1', '1:1.1-1'))
        self.assertFalse(AURDataMapper.check_version_update('1:1.1-1', '1.2-1'))

        self.assertTrue(AURDataMapper.check_version_update('1:1.2-1', '2:0.1-1'))
        self.assertFalse(AURDataMapper.check_version_update('2:0.1-1', '1:1.2-1'))

        self.assertTrue(AURDataMapper.check_version_update('10:1.1-1', '10:1.2-1'))
        self.assertFalse(AURDataMapper.check_version_update('10:1.2-1', '10:1.2-1'))

        self.assertTrue(AURDataMapper.check_version_update('9:1.2-1', '10:0.1-1'))

        self.assertTrue(AURDataMapper.check_version_update('9:1.1.1.1-2', '10:0.0'))
        self.assertFalse(AURDataMapper.check_version_update('10:0.0', '9:1.1.1.1-2'))

    def test_check_update__pkg_no_last_modified_and_same_versions(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.version = '1.0.0'
        pkg.latest_version = pkg.version

        self.assertFalse(mapper.check_update(pkg=pkg, last_modified=1608143812))

    def test_check_update__pkg_no_last_modified_and_latest_version_higher_than_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.version = '1.0.0'
        pkg.latest_version = '1.1.0'

        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=1608143812))

    def test_check_update__pkg_no_last_modified_and_no_install_date_and_version_higher_than_latest_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.install_date = None
        pkg.version = '1.1.0'
        pkg.latest_version = '1.0.0'

        self.assertFalse(mapper.check_update(pkg=pkg, last_modified=1608143812))

    def test_check_update__none_last_modified_and_latest_version_higher_than_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = '1.1.0'

        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=None))

    def test_check_update__none_last_modified_and_version_equal_latest_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = pkg.version

        self.assertFalse(mapper.check_update(pkg=pkg, last_modified=None))

    def test_check_update__string_last_modified_and_latest_version_higher_than_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = '1.1.0'

        self.assertTrue(mapper.check_update(pkg=pkg, last_modified='abc'))

    def test_check_update__pkg_last_modified_equal_last_modified_and_version_equal_latest_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = pkg.version

        self.assertFalse(mapper.check_update(pkg=pkg, last_modified=pkg.last_modified))

    def test_check_update__pkg_last_modified_higher_than_last_modified_and_latest_version_higher_than_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = '1.1.0'

        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=pkg.last_modified - 100))

    def test_check_update__pkg_last_modified_less_than_last_modified_and_version_higher_than_latest_version(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = 1608143812
        pkg.version = '2.0.0'
        pkg.latest_version = '1.0.0'

        # in this case, last modified is more relevant than the string version
        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=pkg.last_modified + 100))

    def test_check_update__pkg_no_last_modified_and_install_date_less_than_last_modified_and_version_higher_than_latest(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.install_date = 1608143812
        pkg.version = '3.0.0'
        pkg.latest_version = '2.0.0'

        # in this case, install_date will be considered instead of package's last_modified.
        # even that 'version' is higher than 'latest_version', 'last_modified' is greater than 'install_date'
        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=pkg.install_date + 100))

    def test_check_update__pkg_no_last_modified_and_install_date_higher_than_last_modified_and_version_equal_latest(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.install_date = 1608143812
        pkg.version = '2.0.0'
        pkg.latest_version = pkg.version

        # in this case, install_date will be considered instead of package's last_modified.
        # as 'install_date' is higher, only the string versions will be compared
        self.assertFalse(mapper.check_update(pkg=pkg, last_modified=pkg.install_date - 100))

    def test_check_update__pkg_no_last_modified_and_install_date_higher_than_last_modified_and_latest_higher(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.install_date = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = '1.1.0'

        # in this case, install_date will be considered instead of package's last_modified.
        # as 'install_date' is higher, only the string versions will be compared
        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=pkg.install_date - 100))

    def test_check_update__pkg_no_last_modified_and_install_date_and_no_last_modified_and_latest_higher(self):
        mapper = AURDataMapper(i18n=Mock(), logger=Mock(), http_client=Mock())
        pkg = ArchPackage(name='test')
        pkg.last_modified = None
        pkg.install_date = 1608143812
        pkg.version = '1.0.0'
        pkg.latest_version = '1.1.0'

        # in this case, install_date will be considered instead of package's last_modified.
        # as 'install_date' is higher, only the string versions will be compared
        self.assertTrue(mapper.check_update(pkg=pkg, last_modified=None))

    def test_check_version_update__one_version_contain_mixed_letters_and_symbols(self):
        self.assertTrue(AURDataMapper.check_version_update('2.2.6', '2.3.3op2'))
        self.assertFalse(AURDataMapper.check_version_update('2.3', '2.2.a'))
        self.assertTrue(AURDataMapper.check_version_update('2.2.a.123', '2.3'))

    def test_check_update__only_installed_version_with_release_number(self):
        self.assertTrue(AURDataMapper.check_version_update('2.2.6-1', '2.3'))
        self.assertTrue(AURDataMapper.check_version_update('2.2', '2.2-2'))
        self.assertFalse(AURDataMapper.check_version_update('2.2', '2.2-1'))

    def test_check_update__versions_with_epoch(self):
        self.assertTrue(AURDataMapper.check_version_update('2.3', '1:2.1'))
        self.assertFalse(AURDataMapper.check_version_update('1:1.0', '2.1'))



from typing import Optional, Set, Iterable, Tuple

from bauh.api.abstract.model import SoftwarePackage, CustomSoftwareAction
from bauh.commons import resource
from bauh.gems.snap import ROOT_DIR


class SnapApplication(SoftwarePackage):

    __actions: Optional[Tuple[CustomSoftwareAction, CustomSoftwareAction]] = None

    @classmethod
    def actions(cls) -> Tuple[CustomSoftwareAction, CustomSoftwareAction]:
        if cls.__actions is None:
            cls.__actions = (
                CustomSoftwareAction(i18n_status_key='snap.action.refresh.status',
                                     i18n_label_key='snap.action.refresh.label',
                                     i18n_description_key='snap.action.refresh.desc',
                                     icon_path=resource.get_path('img/refresh.svg', ROOT_DIR),
                                     manager_method='refresh',
                                     requires_root=True,
                                     i18n_confirm_key='snap.action.refresh.confirm'),
                CustomSoftwareAction(i18n_status_key='snap.action.channel.status',
                                     i18n_label_key='snap.action.channel.label',
                                     i18n_confirm_key='snap.action.channel.confirm',
                                     i18n_description_key='snap.action.channel.desc',
                                     icon_path=resource.get_path('img/refresh.svg', ROOT_DIR),
                                     manager_method='change_channel',
                                     requires_root=True,
                                     requires_confirmation=False)
            )

        return cls.__actions

    def __init__(self, id: str = None, name: str = None, version: str = None, latest_version: str = None,
                 description: str = None, publisher: str = None, rev: str = None, notes: str = None,
                 confinement: str = None, verified_publisher: bool = False,
                 screenshots: Optional[Set[str]] = None,
                 license: Optional[str] = None,
                 installed: bool = False,
                 icon_url: Optional[str] = None,
                 download_size: Optional[int] = None,
                 developer: Optional[str] = None,
                 contact: Optional[str] = None,
                 tracking: Optional[str] = None,
                 app_type: Optional[str] = None,
                 channel: Optional[str] = None,
                 app: bool = False,
                 installed_size: Optional[int] = None):
        super(SnapApplication, self).__init__(id=id, name=name, version=version,
                                              latest_version=latest_version, description=description,
                                              license=license, installed=installed, icon_url=icon_url)
        self.publisher = publisher
        self.rev = rev
        self.notes = notes
        self.confinement = confinement
        self.verified_publisher = verified_publisher
        self.screenshots = screenshots
        self.download_size = download_size
        self.developer = developer
        self.contact = contact
        self.tracking = tracking
        self.type = app_type
        self.channel = channel
        self.app = app
        self.installed_size = installed_size

    def supports_disk_cache(self):
        return self.installed

    def has_history(self):
        return False

    def has_info(self):
        return True

    def can_be_downgraded(self):
        return self.installed

    def get_type(self):
        return 'snap'

    def get_default_icon_path(self):
        return resource.get_path('img/snap.svg', ROOT_DIR)

    def get_type_icon_path(self):
        return self.get_default_icon_path()

    def is_application(self) -> bool:
        if self.installed:
            return self.app
        else:
            return self.type == 'app'

    def get_disk_cache_path(self):
        return f'{super(SnapApplication, self).get_disk_cache_path()}/installed/{self.name}'

    def is_trustable(self) -> bool:
        return self.verified_publisher

    def get_data_to_cache(self):
        return {
            'categories': self.categories
        }

    def fill_cached_data(self, data: dict):
        if data:
            for base_attr in self.get_data_to_cache().keys():
                if data.get(base_attr):
                    setattr(self, base_attr, data[base_attr])

    def can_be_run(self) -> bool:
        return bool(self.installed and self.is_application())

    def get_publisher(self):
        return self.publisher

    def get_custom_actions(self) -> Optional[Iterable[CustomSoftwareAction]]:
        if self.installed:
            return self.actions()

    def supports_backup(self) -> bool:
        return True

    def has_screenshots(self) -> bool:
        return not self.installed and self.screenshots

    def __eq__(self, other):
        if isinstance(other, SnapApplication):
            return self.name == other.name



from typing import Optional, Tuple

from bauh.commons.system import execute


def mkdir(dir_path: str, parent: bool = True, custom_user: Optional[str] = None) -> Tuple[bool, Optional[str]]:
    code, output = execute(f'mkdir {"-p " if parent else ""}"{dir_path}"', shell=True, custom_user=custom_user)
    return code == 0, output



from typing import Optional, Tuple, Collection, Iterable, Generator

from bauh.api.abstract.model import SoftwarePackage, CustomSoftwareAction
from bauh.commons import resource
from bauh.gems.debian import DEBIAN_ICON_PATH, ROOT_DIR


class DebianApplication:
    """
    For packages that represent applications
    """

    def __init__(self, name: str, exe_path: str, icon_path: str, categories: Optional[Tuple] = None):
        self.name = name
        self.exe_path = exe_path
        self.icon_path = icon_path
        self.categories = categories

    def __eq__(self, other):
        if isinstance(other, DebianApplication):
            return self.__dict__ == other.__dict__

        return False

    def __hash__(self) -> int:
        hash_sum = 0

        for k, v in self.__dict__.items():
            hash_sum += hash(v)

        return hash_sum

    def __repr__(self):
        return f"{self.__class__.__name__} ({', '.join((f'{k}={v}' for k, v in self.__dict__.items()))})"

    def to_index(self) -> dict:
        return {self.name: {f: v for f, v in self.__dict__.items() if f != 'name'}}


class DebianPackage(SoftwarePackage):

    __actions: Optional[Tuple[CustomSoftwareAction]] = None

    @classmethod
    def custom_actions(cls) -> Tuple[CustomSoftwareAction]:
        if cls.__actions is None:
            cls.__actions = ((CustomSoftwareAction(i18n_label_key='debian.action.purge',
                                                   i18n_status_key='debian.action.purge.status',
                                                   i18n_description_key='debian.action.purge.desc',
                                                   icon_path=resource.get_path('img/clean.svg', ROOT_DIR),
                                                   manager_method='purge',
                                                   requires_root=True,
                                                   requires_confirmation=False),))

        return cls.__actions

    def __init__(self, name: str = None, version: Optional[str] = None, latest_version: Optional[str] = None,
                 description: Optional[str] = None, maintainer: Optional[str] = None, installed: bool = False,
                 update: bool = False, app: Optional[DebianApplication] = None, compressed_size: Optional[int] = None,
                 uncompressed_size: Optional[int] = None, categories: Tuple[str] = None,
                 updates_ignored: Optional[bool] = None, transaction_size: Optional[int] = None):
        super(DebianPackage, self).__init__(id=name, name=name, version=version, installed=installed,
                                            description=description, update=update,
                                            latest_version=latest_version if latest_version is not None else version)
        self.maintainer = maintainer
        self.compressed_size = compressed_size
        self.uncompressed_size = uncompressed_size
        self.categories = categories
        self.app = app
        self.bind_app(app)
        self.updates_ignored = updates_ignored
        self.transaction_size = transaction_size  # size in bytes related to a transaction (install, upgrade, remove)

    def bind_app(self, app: Optional[DebianApplication]):
        self.app = app

        if app and app.categories:
            self.categories = app.categories

    def has_history(self) -> bool:
        return False

    def has_screenshots(self) -> bool:
        return False

    def has_info(self) -> bool:
        return True

    def can_be_downgraded(self) -> bool:
        return False

    def get_type(self):
        return 'debian'

    def get_default_icon_path(self) -> str:
        return self.get_type_icon_path()

    def get_type_icon_path(self) -> str:
        return DEBIAN_ICON_PATH

    def is_application(self):
        return bool(self.app)

    def get_data_to_cache(self) -> dict:
        pass

    def fill_cached_data(self, data: dict):
        pass

    def can_be_run(self) -> bool:
        return bool(self.app)

    def get_publisher(self) -> str:
        return self.maintainer

    def supports_backup(self) -> bool:
        return True

    def get_disk_icon_path(self) -> Optional[str]:
        if self.app:
            return self.app.icon_path

    def get_custom_actions(self) -> Optional[Iterable[CustomSoftwareAction]]:
        if self.installed:
            return self.custom_actions()

    def is_update_ignored(self) -> bool:
        return bool(self.updates_ignored)

    def supports_ignored_updates(self) -> bool:
        return True

    def __eq__(self, other):
        if isinstance(other, DebianPackage):
            return self.name == other.name

        return False

    def __hash__(self):
        return hash(self.name)

    def __repr__(self) -> str:
        attrs = ', '.join((f'{p}={v}' for p, v in sorted(self.__dict__.items())))
        return f"{self.__class__.__name__} ({attrs})"


class DebianTransaction:

    def __init__(self, to_install: Optional[Collection[DebianPackage]],
                 to_remove: Optional[Collection[DebianPackage]],
                 to_upgrade: Optional[Collection[DebianPackage]]):

        self.to_install = to_install
        self.to_remove = to_remove
        self.to_upgrade = to_upgrade

    @property
    def all_packages(self) -> Generator[DebianPackage, None, None]:
        for pkgs in (self.to_install, self.to_remove, self.to_upgrade):
            if pkgs:
                yield from pkgs

    def __eq__(self, other) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, DebianTransaction) else False

    def __hash__(self) -> int:
        return sum((hash(v)for v in self.__dict__.values()))



import os
import shutil
from typing import List, Optional

from bauh.commons.system import SimpleProcess, run_cmd
from bauh.gems.web import NATIVEFIER_BIN_PATH, NODE_PATHS, ELECTRON_CACHE_DIR


def install(url: str, name: str, output_dir: str, electron_version: Optional[str], cwd: str, system: bool,
            user_agent: Optional[str] = None, extra_options: List[str] = None) -> SimpleProcess:
    cmd = [NATIVEFIER_BIN_PATH if not system else 'nativefier', url, '--name', name, output_dir]

    if electron_version:
        cmd.append('-e')
        cmd.append(electron_version)

    if user_agent:
        cmd.extend(('--user-agent', user_agent))

    if extra_options:
        cmd.extend(extra_options)

    return SimpleProcess(cmd, cwd=cwd, extra_paths=NODE_PATHS if not system else None,
                         extra_env={'XDG_CACHE_HOME': os.path.dirname(ELECTRON_CACHE_DIR)})


def is_available() -> bool:
    return bool(shutil.which('nativefier'))


def get_version() -> str:
    return run_cmd('{} --version'.format(NATIVEFIER_BIN_PATH), print_error=False, extra_paths=NODE_PATHS)


